import 'package:boost/boost.dart';
import 'package:cl_appbase/bloc.dart';
import 'package:flutter/material.dart';

import 'collection_builder.dart';

class SliverCollectionBuilder<TValue> extends StatefulWidget {
  final pullToRefresh;
  final ScrollPhysics? physics;
  final Axis scrollDirection;
  final ScrollBehavior? scrollBehavior;

  final ReactiveCollection<TValue> collection;
  final List<Widget>? leading;
  final Widget Function(BuildContext context, TValue) itemBuilder;
  final Widget Function(BuildContext context)? loadingBuilder;
  final Widget Function(BuildContext context)? emptyBuilder;
  final Widget Function(BuildContext context, dynamic e)? errorBuilder;

  const SliverCollectionBuilder({
    Key? key,
    required this.collection,
    required this.itemBuilder,
    this.leading,
    this.errorBuilder,
    this.emptyBuilder,
    this.loadingBuilder,
    this.physics,
    this.scrollDirection = Axis.vertical,
    this.scrollBehavior,
    this.pullToRefresh = false,
  }) : super(key: key);

  @override
  _SliverCollectionBuilderState<TValue> createState() =>
      _SliverCollectionBuilderState<TValue>();
}

class _SliverCollectionBuilderState<TValue>
    extends State<SliverCollectionBuilder<TValue>> {
  final _continueSemaphore = Semaphore();
  bool _done = false;
  final key = GlobalKey();

  @override
  void initState() {
    super.initState();
    _continueLoading();
  }

  @override
  Widget build(BuildContext context) {
    return CollectionBuilder<TValue>(
      key: key,
      stream: widget.collection.stream,
      builder: _build,
      error: widget.errorBuilder ?? _fallbackError,
      empty: widget.emptyBuilder,
    );
  }

  Widget _build(BuildContext context, List<TValue> items, bool loading) {
    if (widget.pullToRefresh) {
      return RefreshIndicator(
        child: _buildScrollView(context, items, loading),
        onRefresh: _refresh,
      );
    } else {
      return _buildScrollView(context, items, loading);
    }
  }

  Widget _buildScrollView(
      BuildContext context, List<TValue> items, bool loading) {
    return NotificationListener<ScrollNotification>(
      onNotification: _onScroll,
      child: CustomScrollView(
        physics: widget.physics,
        scrollDirection: widget.scrollDirection,
        scrollBehavior: widget.scrollBehavior,
        slivers: [
          if (widget.leading != null) ...widget.leading!,
          ...items.map((e) => widget.itemBuilder(context, e)),
          if (loading)
            widget.loadingBuilder?.call(context) ??
                SliverToBoxAdapter(
                  child: SizedBox(
                    height: 64.0,
                    child: Center(
                      child: SizedBox(
                        child: SizedBox(
                          height: 32.0,
                          width: 32.0,
                          child: CircularProgressIndicator.adaptive(),
                        ),
                      ),
                    ),
                  ),
                )
        ],
      ),
    );
  }

  bool _onScroll(ScrollNotification notification) {
    if (!_done &&
        notification.metrics.pixels >
            notification.metrics.maxScrollExtent - 300) {
      _continueLoading();
    }
    return false; //don't intercept bubbling
  }

  Future<void> _continueLoading({bool refresh = false}) async {
    await _continueSemaphore.throttle(() async {
      _done = await widget.collection.load(invalidate: refresh);
    });
  }

  Widget _fallbackError(BuildContext context, dynamic e) {
    return Text(e.toString());
  }

  Future<void> _refresh() async {
    await _continueLoading(refresh: true);
  }
}
